import sys



from Bio import AlignIO

sys.path.append("..")
#from recan.simgen import Simgen

from recan.calc_pairwise_distance import p_distance, estimate_nucleotide_frequencies
from recan.rolling_window import RollingWindowOnAlignment


def test_p_distance():
    
    align = AlignIO.read("test_p_dist.fasta", "fasta")
    """
    >s1_ref
    ATGCATGCAT
    >s2_dist_0
    ATGCATGCAT
    >s3_dist_10
    TTGCATGCAT
    >s4_dist_50
    GGCGGTGCAT
    >s5_dist_90
    GGCGGGGGGG
    >s5_dist_100
    GGCGGGCGGG
    """
    ref_seq = align[0]
    seq_0 = align[1]
    seq_10 = align[2]
    seq_50 = align[3]
    seq_90 = align[4]
    seq_100 = align[5]
    
    assert p_distance(ref_seq, seq_0) == 0.0
    assert p_distance(ref_seq, seq_10) == 0.1
    assert p_distance(ref_seq, seq_50) == 0.5
    assert p_distance(ref_seq, seq_90) == 0.9
    assert p_distance(ref_seq, seq_100) == 1.0
    

def test_estimate_nuc_frequency():
    align = AlignIO.read("test_nuc_freq.fasta", "fasta")
    
    """ sequences in the file 'test_nuc_freq.fastq":
    >s1_A_10
    ATGCCTGCTT
    >s2_G_50
    AGCGTGCGTG
    >s3_C_90
    CCCCACCCCC
    >s4_T_30
    TAGCTAGCTA
    >s5_A_0
    TGCTTGCTTG
    >s5_G_100
    GGGGGGGGGG
    """
    # return [ x / length for x in [A, C ,G, T] ]
    
    # seq to get Seq obj out of SeqRecord obj
    assert estimate_nucleotide_frequencies(align[0].seq)[0] == 0.1
    assert estimate_nucleotide_frequencies(align[1].seq)[2] == 0.5
    assert estimate_nucleotide_frequencies(align[2].seq)[1] == 0.9
    assert estimate_nucleotide_frequencies(align[3].seq)[3] == 0.3
    assert estimate_nucleotide_frequencies(align[4].seq)[0] == 0.0
    assert estimate_nucleotide_frequencies(align[5].seq)[2] == 1.0
    

def test_sliced_alignment_slices_on_whole_alignment():
    
    align = RollingWindowOnAlignment("./hbv_C_Bj_Ba.fasta")
    
    sliced_align = align.roll_window_along_alignment(window_len=500, window_step=500)
    assert len(sliced_align) == 7
    
    sliced_align = align.roll_window_along_alignment(window_len=1000, window_step=500)
    assert len(sliced_align) == 7
    
    sliced_align = align.roll_window_along_alignment(window_len=1, window_step=1)
    assert len(sliced_align) == 3215
    
    sliced_align = align.roll_window_along_alignment(window_len=3215, window_step=3215)
    assert len(sliced_align) == 1
    

def test_sliced_alignment_slices_on_alignment_region():
    
    align = RollingWindowOnAlignment("./hbv_C_Bj_Ba.fasta")
    
    sliced_align = align.roll_window_along_alignment_region(window_len=500, window_step=500, 
                                                     region=(0, 1000))
    assert len(sliced_align) == 2
    
    sliced_align = align.roll_window_along_alignment_region(window_len=500, window_step=250, 
                                                     region=(0, 1000))
    assert len(sliced_align) == 4
    
    sliced_align = align.roll_window_along_alignment_region(window_len=500, window_step=250, 
                                                     region=(0, 3215))
    assert len(sliced_align) == 13
    
    
def test_sliced_alignment_window_borders_whole_alignment():
    
    align = RollingWindowOnAlignment("./hbv_C_Bj_Ba.fasta")
    
    sliced_align = align.roll_window_along_alignment(window_len=500, window_step=500)
    
    
    check_window_coords = [[0, 500], [500, 1000], [1000, 1500], [1500, 2000],
                     [2000, 2500], [2500, 3000], [3000, 3215]]
    
    counter = 0
    for window_coords in sliced_align.keys():
        
        assert window_coords[0] == check_window_coords[counter][0]
        assert window_coords[1] == check_window_coords[counter][1]
        counter += 1
        

   
def test_sliced_alignment_window_borders_alignment_region():
    
    align = RollingWindowOnAlignment("./hbv_C_Bj_Ba.fasta")
    
    sliced_align = align.roll_window_along_alignment_region(window_len=500, window_step=250, 
                                                     region=(0, 1000))
    
    
    check_window_coords = [[0, 500], [250, 750], [500, 1000], [750, 1000]]
    
    counter = 0
    for window_coords in sliced_align.keys():
        
        assert window_coords[0] == check_window_coords[counter][0]
        assert window_coords[1] == check_window_coords[counter][1]
        counter += 1
        
       
    
    
    
    
    
    
    




# -*- coding: utf-8 -*-

import plotly.graph_objs as go
import pandas as pd
from plotly.offline import init_notebook_mode, iplot
from .rolling_window import RollingWindowOnAlignment
from .calc_pairwise_distance import calc_pairwise_distance


class Simgen():
    
    def __init__(self, in_file):
        
        self.alignment_roll_window = RollingWindowOnAlignment(in_file)
        self.align_sliced = None
        self.ticks_for_x_axis = []
        self.distance_data = {}
        self.pot_rec_index = None
        self.pot_rec_id = None
        
        
    def _plot_distance_by_plotly(self):
                
        """draws similarity plot using plotly"""
        init_notebook_mode()
        data = []
        for key in self.distance_data.keys():
            trace = go.Scatter(y=self.distance_data[key], x=self.ticks_for_x_axis, name=key)
            data.append(trace)
        
        layout = go.Layout(
            xaxis=dict(
                title="nucleotide position"),
            yaxis=dict(
                title="sequence identity"),
            legend=dict(x=-0.1, y=1.5, orientation="h"))
            #legend=dict(x=-0.1, y=1.5))
            
        fig = go.Figure(data=data, layout=layout)
        iplot(fig)
        print("potential recombinant: ", self.pot_rec_id)
        
        
    def _get_ticks_for_x_axis(self):
        """
        prepares ticks for x axis
        the ticks correspond to the alignment nucleotide positions
        the ticks for x axis are in the middle of the rolling window
        """
        
        self.ticks_for_x_axis.clear()
        
        for start_stop_nuc_index in self.align_sliced.keys():
            # we take a nucleotide position in the middle of the window
            # it'll be a point we plot
            middle_nuc = (start_stop_nuc_index[1] - start_stop_nuc_index[0]) / 2
            self.ticks_for_x_axis.append(start_stop_nuc_index[0] + middle_nuc)
            
    
    def _get_pot_rec_id(self):
        """ get id of the potential recombinant from
        the alignment slices
        """
        
        for slice in self.align_sliced.values():
            slice_0 = slice
            break
        
        self.pot_rec_id = slice_0[self.pot_rec_index].id
        
    
    
    def _prepare_distance_data(self):
        
        """ makes dictionary for distance data collection
        keys are sequences names (ids from the alignment)
        values are lists of distances to the potential recombinant
        """
        self.distance_data = {}
        
        for slice in self.align_sliced.values():
            slice_0 = slice
            break
        
        for seq in slice_0:
            if seq.id == self.pot_rec_id:
                continue
            self.distance_data[seq.id] = []
            
    
    
    def simgen(self, pot_rec, window, shift, region=False, dist="pdist"):
        """
        Parameters
        ----------
        pot_rec : int
            index of the potential recombinant in the alignment.
        window : int
            sliding window size.
        shift : int
            sliding window shift along the alignment.
        region : list or tuple, optional
             The default is False. start and end of the region
             to analyze, f.e. (1000, 3000)
        dist : str, optional
            the distance calculation method. default is "pdist". 
            
            available methods:
            pdist - pairwise distance (default)
            jcd - Jukes-Cantor distance
            k2p -  Kimura 2-parameter distance
            td - Tamura distance
            
        Returns
        -------
        None.

        """
        
        if region:
            assert region[0] < region[1], "start of the region must be less than the region end"
            self.align_sliced = self.alignment_roll_window.roll_window_along_alignment_region(window_len=window, 
                                                                          window_step=shift,
                                                                          region=region)
        else:
            self.align_sliced = self.alignment_roll_window.roll_window_along_alignment(window_len=window, 
                                                                          window_step=shift)
        self.pot_rec_index = pot_rec
        self._get_pot_rec_id()
        self._get_ticks_for_x_axis()
        self._prepare_distance_data()
        
        
        for window_borders, alignment_slice in self.align_sliced.items():
            for seq in alignment_slice:
                if seq.id == self.pot_rec_id:
                    continue
                seq1 = alignment_slice[self.pot_rec_index].seq
                seq2 = seq.seq
                distance = calc_pairwise_distance(seq1=seq1, seq2=seq2, 
                                                 dist_method=dist) 
                self.distance_data[seq.id].append(distance)

        
        self._plot_distance_by_plotly()
        
        
        
        
    def save_data(self, path=False, out="csv", out_name="distance_data",
                  data_cols="plot_ticks"):
        """saves the data spreadsheet as a csv file
        Parameters
        ---------
        path: str
            output destination
        out: str
            output file format: "csv" or "excel"
        out_name: str
            output file name
        columns: str
            denoting columns in the spreadsheet: "plot_ticks" by default,
            or "window_pos" for sliding window start and end nucleotide 
            position in alignment
        """
        
        if data_cols == "plot_ticks":
            columns = self.ticks_for_x_axis
        elif data_cols == "window_pos":
            columns = [x for x in self.align_sliced.keys()]
       
        
        df = pd.DataFrame.from_dict(self.distance_data, 
                                    orient='index',
                                    columns=columns)
       
        if path:
            if out == "csv":
                df.to_csv(out_name + ".csv")
            else:
                print("invalid output file")
        else:
            if out == "csv":
                df.to_csv(out_name + ".csv")
            else:
                print("invalid output file format")
        
   
        
        
    def get_data(self, df=True):
        """returns distance data
        Parameters
        ---------
        df: bool
            True: returns pandas DataFrame object
            False: returns a dictionary where keys are the sequence ids and 
            values are distance data
        
        """
        if df:
            return pd.DataFrame(data=self.distance_data, index=self.ticks_for_x_axis).T
        else:
            return self.ticks_for_x_axis, self.distance_data
        
        
    def get_info(self):
        """outputs information about the alignment: 
        index (which is the row number), 
        sequence names, and alignment length"""
        
        print("index:", "sequence id:", sep="\t")
        for seq_index, seq in enumerate(self.alignment_roll_window.align):
            print(seq_index, seq.id, sep="\t")
        print("alignment length: ", self.alignment_roll_window.align.get_alignment_length())
      
        
        
    


# -*- coding: utf-8 -*-


from math import log, sqrt

"""
U	Uracil (RNA)	U
W	Weak	A/T
S	Strong	C/G
M	Amino	A/C
K	Keto	G/T
R	Purine	A/G
Y	Pyrimidine	C/T
B	Not A	C/G/T
D	Not C	A/G/T
H	Not G	A/C/T
V	Not T	A/C/G
N	Any	A/C/G/T
"""
DEGENERATE_NUCS = ("U", "W", "S", "M", "K", "R", "Y", "B", "D", "H", "V", "N")

def estimate_nucleotide_frequencies(seq):
    
    seq = seq.replace("-","").upper()
    
    A = seq.count("A")
    C = seq.count("C")
    G = seq.count("G")
    T = seq.count("T")
    
    length = float(len(seq))
    
    return [ x / length for x in [A, C ,G, T] ]
    
    
def p_distance(seq1, seq2):
    """calculates pairwise distance between two sequences
    distance = num of different nucleotides / total nucleotides
    """
    
    different_nucs = 0
    nuc_pairs = []
    
    for x in zip(seq1, seq2):
        if '-' not in x and x[0] not in DEGENERATE_NUCS and x[1] not in DEGENERATE_NUCS: # skip gaps
            nuc_pairs.append(x)
            
    for (nuc_1, nuc_2) in nuc_pairs:
        if nuc_1 != nuc_2:
            different_nucs += 1
    
    total_nucs = len(nuc_pairs)
    
    try:
        distance = float(different_nucs / total_nucs)
        
        return distance
    
    except ValueError:
        print("""the reasons for the ValueError maybe: 1. too many gaps in some
              region of the alignment. 2. too short window span""")
    
        
    
    
def jc_distance(seq1, seq2):
    
    """ 
    Jukes-Cantor 
    jc_distance = - b log(1 - p_dist / b)
    ------------
    b is a constant. b = 3/4 for nucleotides and 19/20 for proteins.
    p_dist = pairwise distance, which is uncorrected distance between seq1 and seq2
    """
    b = 0.75
    p_dist = p_distance(seq1,seq2)
    
    try: 
        distance = - b * log(1 - p_dist / b)
        
        return distance
    
    except ValueError: 
        print("""the reasons for the ValueError maybe: 1. too many gaps in some
              region of the alignment. 2. too short window span""")
        

       
def k2p_distance(seq1,seq2):
    """
    Kimura 2-Parameter distance 
    k2p_distance = - 0.5 log( (1 - 2p - q)*sqrt(1 - 2q) )
    where:
    p : transition frequency
    q : transversion frequency
    """
    nuc_pairs = []

    # collect nuc pairs without gaps
    for x in zip(seq1,seq2):
        if '-' not in x and x[0] not in DEGENERATE_NUCS and x[1] not in DEGENERATE_NUCS: 
            nuc_pairs.append(x)
        
    ts_count = 0
    tv_count = 0
    total_nucs = len(nuc_pairs)
    
    transitions = [ "AG", "GA", "CT", "TC"]
    transversions = [ "AC", "CA", "AT", "TA",
                      "GC", "CG", "GT", "TG" ]

    for (nuc_1, nuc_2) in nuc_pairs:
        if nuc_1 + nuc_2 in transitions: 
            ts_count += 1 
        elif nuc_1 + nuc_2 in transversions: 
            tv_count += 1
    
    ts_freq = float(ts_count) / total_nucs
    tv_freq = float(tv_count) / total_nucs
    
    try: 
        distance = -0.5 * log((1 - 2 * ts_freq - tv_freq) * sqrt( 1 - 2 * tv_freq ))
        
        return distance
    
    except ValueError: 
        print("""the reasons for the ValueError maybe: 1. too many gaps in some
              region of the alignment. 2. too short window span""")
        
     
def tamura_distance(seq1,seq2):
    """
    Tamura distance = -C log( 1 - P/C - Q ) - 0.5( 1 - C )log( 1 - 2Q )
    where:
    P = transition frequency
    Q = transversion frequency
    C = GC1 + GC2 - 2 * GC1 * GC2
    GC1 = GC-content of sequence 1
    GC2 = GC-coontent of sequence 2
    """
    
    nuc_pairs = []
    
    #collect ungapped pairs
    for x in zip(seq1,seq2):
        if '-' not in x and x[0] not in DEGENERATE_NUCS and x[1] not in DEGENERATE_NUCS: 
            nuc_pairs.append(x)
        
    ts_count = 0
    tv_count = 0
    total_nucs = len(nuc_pairs)
    
    transitions = [ "AG", "GA", "CT", "TC"]
    transversions = [ "AC", "CA", "AT", "TA",
                      "GC", "CG", "GT", "TG" ]

    for (nuc_1, nuc_2) in nuc_pairs:
        if nuc_1 + nuc_2 in transitions: 
            ts_count += 1 
        elif nuc_1 + nuc_2 in transversions: 
            tv_count += 1
    
    ts_freq = float(ts_count) / total_nucs # p
    tv_freq = float(tv_count) / total_nucs # q
    
    gc1 = sum(estimate_nucleotide_frequencies(seq1)[1:3])
    gc2 = sum(estimate_nucleotide_frequencies(seq2)[1:3])
    
    c = gc1 + gc2 - 2 * gc1 * gc2

    try: 
        distance = -c * log( 1 - ts_freq / c - tv_freq) - 0.5 * ( 1 - c ) * log ( 1 - 2 * tv_freq )
        
        return distance
    
    except ValueError: 
        print("""the reasons for the ValueError maybe: 1. too many gaps in some
              region of the alignment. 2. too short window span""")
        
    
# TODO tajima nei isn't included still
def tn_distance(seq1, seq2):
    """ 
    Tajima-Nei distance = -b log(1 - p / b)
    where:
    b = 0.5 * [ 1 - Sum i from A to T(Gi^2+p^2/h) ]
    h = Sum i from A to G( Sum j from C to T (Xij^2/2*Gi*Gj))
    p = p-distance, i.e. uncorrected distance between seq1 and seq2
    Xij = frequency of pair (i,j) in seq1 and seq2, with gaps removed
    Gi = frequency of base i over seq1 and seq2 """
    from math import log

    ns = ['A','C','G','T']
    G = estimate_nucleotide_frequencies(seq1 + seq2)
    p = p_distance(seq1,seq2)
    pairs = []
    h = 0

    #collect ungapped pairs
    for x in zip(seq1,seq2):
        if '-' not in x: pairs.append(x)
       
    #pair frequencies are calculated for AC, AG, AT, CG, CT, GT (and reverse order)
    for i in range(len(ns)-1):
        for j in range(i+1,len(ns)):
            if i != j: 
                paircount = pairs.count( (ns[i], ns[j]) ) + pairs.count( (ns[j], ns[i]) )
            Xij_sq = (float(paircount)/len(pairs))**2
            GiGj = G[i]*G[j]
            h += 0.5*Xij_sq/GiGj  #h value used to calculate b
    
    b = 0.5*(1-sum([x**2 for x in G])+p**2/h)
    
    try: 
        d = -b * log(1 - p/b)
        return d 
    
    except ValueError: 
        print("""the reasons for the ValueError maybe: 1. too many gaps in some
              region of the alignment. 2. too short window span""")
        
       
def calc_pairwise_distance(seq1, seq2, dist_method):
    
    if dist_method == "pdist":
        distance = p_distance(seq1, seq2)
        
    elif dist_method == "jcd":
        distance = jc_distance(seq1, seq2)
        
    elif dist_method == "k2p":
        distance = k2p_distance(seq1, seq2)
        
    elif dist_method == "td":
        distance = tamura_distance(seq1, seq2)
    
    #elif dist_method == "tnd":
    #    distance = tn_distance(seq1, seq2)
           
    return 1 - distance







from Bio import AlignIO

class RollingWindowOnAlignment():
    """
    alignment obj as the biopython multiple alignment
    has two sliding window methods that slice the alignment into sections
    """
    
    def __init__(self, in_file):
        self.align = AlignIO.read(in_file, "fasta")
        
    def roll_window_along_alignment(self, window_len, window_step):
     
        window_start = 0
        window_end = window_len
        window_step = window_step

        window_counter = 0
        sliced_alignment = {}
        
        while window_start < self.align.get_alignment_length():
            
            sliced_alignment[(window_start, window_end)] = self.align[:, window_start:window_end]
            window_start += window_step
            
            ###############################################
            # redo
            #new_window_end = window_end + window_step
            if window_end + window_step < self.align.get_alignment_length():
                window_end += window_step
            else:
                window_end = self.align.get_alignment_length()
            # redo
            ############################################3
            window_counter += 1

        return sliced_alignment 
    
    
    def roll_window_along_alignment_region(self, window_len, window_step, region):
    
        
        window_start = region[0]
        window_end = region[0] + window_len
        window_step = window_step

        window_counter = 0
        sliced_alignment = {}
        while window_start < region[1]:
            sliced_alignment[(window_start, window_end)] = self.align[:, window_start:window_end]
            window_start += window_step
            
            if window_end + window_step < region[1]:
                window_end += window_step
            else:
                window_end = region[1]
            

            window_counter += 1
            

        return sliced_alignment
    
    
    
    
    


from os import path
from distutils.core import setup
import setuptools 


this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


   
setup(
  name = 'recan',
  long_description = long_description,  # added to package readme on pypi
  long_description_content_type = "text/markdown",  # added to package readme on pypi
  packages = ['recan'],   
  version = '0.1.4',      
  license='MIT',        
  description = 'recan: recombination analysis tool',   
  author = 'Yuriy Babin',                  
  author_email = 'babin.yurii@gmail.com',      
  url = 'https://github.com/babinyurii/recan', 
  download_url = 'https://github.com/babinyurii/recan/archive/refs/tags/v_0.1.4.tar.gz',
  keywords = ['DNA recombination', 'bioinformatics', 'genetic distance'],   
  install_requires=[            
          'pandas',
          'plotly',
          'biopython',
          'matplotlib'
      ],
  classifiers=[
    'Development Status :: 4 - Beta',      
    'Intended Audience :: Developers',      
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',   
    'Programming Language :: Python :: 3',      
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
)
